/*
 * This program uses a simple mailbox-based protocol to provide RSA encryption
 * computation to a client process.
 *
 * Author:	David Jones
 * Date:	19-JUN-1998
 */
#include <stdio.h>
#include <stdlib.h>
#include <descrip.h>			/* VMS string descriptors */
#include <jpidef.h>
#include <lckdef.h>
#include <iodef.h>
#include <ssdef.h>
#include <string.h>

#include "sshrsa.h"
#include "sshrsa_rpc.h"
struct mbx_iosb {
   unsigned short status, length;
   long pid;
};

static short mbx_chan;
int SYS$ASSIGN(), SYS$QIOW(), LIB$GETJPI(), SYS$ENQ(), SYS$EXIT(), SYS$SETPRN();
int SYS$SETPRI();
static $DESCRIPTOR(engine_name,"RSA engine");

#ifdef DEBUG
static void dump_buffer ( char *heading, char *buffer, int bufsize )
{
    int i, j;
    printf ( "%s (%d bytes):", heading, bufsize );
    j = 15;
    for ( i = 0; i < bufsize; i++ ) {
	j++;
	if ( j > 15 ) {
	    printf("\n   %04x:", i ); j = 0; }
	printf(" %02x", 255&buffer[i] );
    }
    printf("\n");
}
#else
#define dump_buffer(a,b,c) 1
#endif
/***********************************************************************/

int main ( int argc, char **argv )
{
    int status, request_dlen, reply_dlen;
    long pid, code, lksb[2];
    struct mbx_iosb iosb;
    struct mbx_message request, reply;
    static char lock_name_buf[64];
    static $DESCRIPTOR(lock_name,lock_name_buf);
    /*
     * Get PID of owner process, any traffic to mailbox must be to/from this
     * process.  Get our own PID and queue lock  If we get lock, then
     * parent image has exitted.
     */
    status = SYS$SETPRI ( 0, 0, 4, 0, 0, 0 );
    code = JPI$_OWNER;
    status = LIB$GETJPI ( &code, 0, 0, &pid, 0 );
    status = SYS$SETPRN ( &engine_name );
    sprintf ( lock_name_buf, "%s_%x",mbx_name.dsc$a_pointer, pid );
    lock_name.dsc$w_length = strlen ( lock_name.dsc$a_pointer );
    status = SYS$ENQ ( 9, LCK$K_EXMODE, lksb, LCK$M_NODLCKWT, &lock_name,
	0, SYS$EXIT, SS$_ABORT,  0, 0, 0, 0, 0 );
    /*
     * Initialize random number genrator and setup RSA calls for local use.
     */
    sshrsa_init("");
    /*
     * Assign channel to mailbox.
     */
    status = SYS$ASSIGN ( &mbx_name, &mbx_chan, 0, 0, 0 );
    /*
     * Enter main loop: read message and process.
     */
    while ( status&1 ) {
	/*
	 * read request.
	 */
	status = SYS$QIOW ( 3, mbx_chan, IO$_READVBLK, &iosb, 0, 0,
		&request, sizeof(request), 0, 0, 0, 0 );
	/* printf("Message from mailbox sts=%d %d %d %x: code=%d, id=%x\n",
		status, iosb.status, iosb.length, iosb.pid, request.code,
		request.key_id ); */
	if ( status&1 ) status = iosb.status;
	if ( (status&1) == 0 ) break;
	if ( iosb.pid != pid ) continue;
	/*
	 * Setup reply
	 */
	reply_dlen = 0;
	reply.code = request.code;
	reply.key_id = (void *) 0;
	reply.len1 = 0;
	reply.len2 = 0;
        /*
	 * Peform requested operation.
	 */
	request_dlen = iosb.length - (sizeof(request)-sizeof(request.data));
	if ( request_dlen < 0 ) request_dlen = 0;
	status = sshrsa_remote_request 
		(&request, request_dlen, &reply, &reply_dlen);
	/*
	 * Send final reply, verify recipient.
	 */
	/* printf("reply length: %d code: %d, id: %x\n", reply_dlen, 
		reply.code, reply.key_id ); */
	if ( reply_dlen > 0 ) dump_buffer ( "reply data", reply.data,reply_dlen);
	status = SYS$QIOW ( 3, mbx_chan, IO$_WRITEVBLK, &iosb, 0, 0,
		&reply, (sizeof(reply)-sizeof(reply.data)+reply_dlen),
		0, 0, 0, 0 );
	/* printf("reply send status: %d %d %d %x\n", status, iosb.status,
		iosb.length, iosb.pid ); */
	if ( (status&1) == 1 ) status = iosb.status;
    }
    return status;
}
